@@ -42,8 +42,27 @@
 
 namespace firebuild {
 
+/* RAII helper to automatically free inline data on scope exit */
+class InlineDataMapCleaner {
+ public:
+  explicit InlineDataMapCleaner(tsl::hopscotch_map<const FileName*, std::pair<char*, size_t>>* map)
+      : map_(map) {}
+  ~InlineDataMapCleaner() {
+    if (map_) {
+      for (auto& entry : *map_) {
+        free(entry.second.first);
+      }
+    }
+  }
+  /* Prevent copying */
+  InlineDataMapCleaner(const InlineDataMapCleaner&) = delete;
+  InlineDataMapCleaner& operator=(const InlineDataMapCleaner&) = delete;
+ private:
+  tsl::hopscotch_map<const FileName*, std::pair<char*, size_t>>* map_;
+};
+
 static const XXH64_hash_t kFingerprintVersion = 0;
-static const unsigned int kCacheFormatVersion = 1;
+static const unsigned int kCacheFormatVersion = 2;
 static const char kCacheStatsFile[] = "stats";
 static const char kCacheSizeFile[] = "size";
 
@@ -503,20 +522,26 @@ void ExecedProcessCacher::erase_fingerprint(const ExecedProcess *proc) {
 }
 
 static void add_file(std::vector<FBBSTORE_Builder_file>* files, const FileName* file_name,
-                     const FileInfo& fi) {
+                     const FileInfo& fi, const char* inline_data = nullptr,
+                     size_t inline_data_len = 0) {
   FBBSTORE_Builder_file& new_file = files->emplace_back();
   new_file.set_path_with_length(file_name->c_str(), file_name->length());
   new_file.set_type(fi.type());
   if (fi.size_known()) {
     new_file.set_size(fi.size());
   }
-  if (fi.hash_known()) {
+  /* If inline data is provided, omit the hash in the serialized entry to save space.
+   * For non-inlined files, keep storing the hash so they can be restored from the blob cache. */
+  if (!inline_data && fi.hash_known()) {
     new_file.set_hash(fi.hash().get());
   }
   if (fi.mode_mask() != 0) {
     new_file.set_mode(fi.mode());
     new_file.set_mode_mask(fi.mode_mask());
   }
+  if (inline_data && inline_data_len > 0) {
+    new_file.set_inline_data(inline_data, inline_data_len);
+  }
 }
 
 static const FBBSTORE_Builder* file_item_fn(int idx, const void *user_data) {
@@ -666,21 +691,6 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
     if (parent_may_be_just_sh_c_this) {
       joined_cmdline += (joined_cmdline == "") ? arg : (" " + arg);
     }
-    if (arg == "-emit-pch") {
-      bool fno_pch_timestamp_found = false;
-      for (const auto& arg_inner_loop : args) {
-        if (arg_inner_loop == "-fno-pch-timestamp") {
-          fno_pch_timestamp_found = true;
-          break;
-        }
-      }
-      if (!fno_pch_timestamp_found) {
-        proc->disable_shortcutting_bubble_up(
-            "Clang's -emit-pch without -Xclang -fno-pch-timestamp prevents shortcutting");
-        return;
-      }
-      break;
-    }
   }
 
   if (parent_may_be_just_sh_c_this && joined_cmdline == parent_exec_point->args()[2]) {
@@ -769,6 +779,10 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
   }
 
   off_t stored_blob_bytes = 0;
+  /* Map to store inline data temporarily for files */
+  tsl::hopscotch_map<const FileName*, std::pair<char*, size_t>> inline_data_map;
+  InlineDataMapCleaner inline_data_cleaner(&inline_data_map);
+
   for (const auto& pair : proc->file_usages()) {
     const auto filename = pair.first;
     const FileUsage* fu = pair.second;
@@ -798,11 +812,17 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
           int fd = open(filename->c_str(), O_RDONLY);
           if (fd >= 0) {
             off_t stored_bytes = 0;
-            if (!hash_cache->store_and_get_hash(filename, 0, &new_hash, &stored_bytes, fd, &st)) {
+            char *inline_data = nullptr;
+            size_t inline_data_len = 0;
+            if (!hash_cache->store_and_get_hash(filename, 0, &new_hash, &stored_bytes, fd, &st,
+                                               &inline_data, &inline_data_len)) {
               /* unexpected error, now what? */
               FB_DEBUG(FB_DEBUG_CACHING,
                        "Could not store blob in cache, not writing shortcut info");
               close(fd);
+              if (inline_data) {
+                free(inline_data);
+              }
               proc->disable_shortcutting_only_this(
                   "Could not store blob in cache, not writing shortcut info");
               return;
@@ -810,6 +830,10 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
             close(fd);
             new_file_info.set_size(st.st_size);
             new_file_info.set_hash(new_hash);
+            /* Store inline data temporarily for later use in add_file */
+            if (inline_data && inline_data_len > 0) {
+              inline_data_map[filename] = {inline_data, inline_data_len};
+            }
             if ((stored_blob_bytes += stored_bytes) > max_entry_size) {
               FB_DEBUG(FB_DEBUG_CACHING,
                        "Could not store blob in cache because it would exceed max_entry_size");
@@ -849,7 +873,14 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
           proc->disable_shortcutting_only_this("Process created a temporary file");
           return;
         }
-        add_file(&out_path_isreg, filename, new_file_info);
+        {
+          auto it = inline_data_map.find(filename);
+          if (it != inline_data_map.end()) {
+            add_file(&out_path_isreg, filename, new_file_info, it->second.first, it->second.second);
+          } else {
+            add_file(&out_path_isreg, filename, new_file_info);
+          }
+        }
         break;
       case ISDIR:
         // FIXME skip adding if the new state is the same as the old one
@@ -890,7 +921,9 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
           bool is_empty;
           Hash hash;
           off_t stored_bytes = 0;
-          if (!recorder->store(&is_empty, &hash, &stored_bytes)) {
+          char *inline_data = NULL;
+          size_t inline_data_len = 0;
+          if (!recorder->store(&is_empty, &hash, &stored_bytes, &inline_data, &inline_data_len)) {
             // FIXME handle error
             FB_DEBUG(FB_DEBUG_CACHING,
                      "Could not store pipe traffic in cache, not writing shortcut info");
@@ -909,7 +942,13 @@ void ExecedProcessCacher::store(ExecedProcess *proc) {
              * They were taken into account when computing the process's fingerprint. */
             FBBSTORE_Builder_append_to_fd& new_append = out_append_to_fd.emplace_back();
             new_append.set_fd(fd);
-            new_append.set_hash(hash.get());
+            if (inline_data) {
+              /* Store inline data in the cache entry */
+              FB_DEBUG(FB_DEBUG_CACHING, "Storing inline data: len=" + d(inline_data_len));
+              new_append.set_inline_data(inline_data, inline_data_len);
+            } else {
+              new_append.set_hash(hash.get());
+            }
           }
         }
       } else if (inherited_file.type == FD_FILE) {
@@ -1225,7 +1264,7 @@ const FBBSTORE_Serialized_process_inputs_outputs * ExecedProcessCacher::find_sho
       if (count == 2) {
         FB_DEBUG(FB_DEBUG_SHORTCUT,
                  "│   More than 1 matching candidates found, still using the first one");
-        munmap(candidate_inouts_buf, candidate_inouts_buf_len);
+        ObjCache::unmap_entry(candidate_inouts_buf, candidate_inouts_buf_len);
         break;
       }
 #else
@@ -1238,7 +1277,7 @@ const FBBSTORE_Serialized_process_inputs_outputs * ExecedProcessCacher::find_sho
       break;
 #endif
     } else {
-      munmap(candidate_inouts_buf, candidate_inouts_buf_len);
+      ObjCache::unmap_entry(candidate_inouts_buf, candidate_inouts_buf_len);
     }
   }
   /* The retval is currently the same as the memory address to unmap (i.e. *inouts_buf).
@@ -1403,8 +1442,8 @@ bool ExecedProcessCacher::apply_shortcut(ExecedProcess *proc,
   for (i = 0; i < outputs->get_path_isreg_count(); i++) {
     auto file = reinterpret_cast<const FBBSTORE_Serialized_file *>(outputs->get_path_isreg_at(i));
     if (file->get_type() == ISREG) {
-      assert(file->has_hash());
-      if (!blob_fds.add_from_hash(file->get_hash())) {
+      /* Skip inline data - it doesn't need blob fd */
+      if ((file->get_inline_data_count() == 0) && !blob_fds.add_from_hash(file->get_hash())) {
         return false;
       }
     }
@@ -1412,7 +1451,9 @@ bool ExecedProcessCacher::apply_shortcut(ExecedProcess *proc,
   for (i = 0; i < outputs->get_append_to_fd_count(); i++) {
     auto append_to_fd = reinterpret_cast<const FBBSTORE_Serialized_append_to_fd *>
         (outputs->get_append_to_fd_at(i));
-    if (!blob_fds.add_from_hash(append_to_fd->get_hash())) {
+    /* Skip inline data - it doesn't need blob fd */
+    if ((append_to_fd->get_inline_data_count() == 0)
+        && !blob_fds.add_from_hash(append_to_fd->get_hash())) {
       return false;
     }
   }
@@ -1446,40 +1487,70 @@ bool ExecedProcessCacher::apply_shortcut(ExecedProcess *proc,
     auto file = reinterpret_cast<const FBBSTORE_Serialized_file *>(outputs->get_path_isreg_at(i));
     const auto path = FileName::Get(file->get_path(), file->get_path_len());
     if (file->get_type() == ISREG) {
-      FB_DEBUG(FB_DEBUG_SHORTCUT,
-               "│   Fetching file from blobs cache: "
-               + d(path));
-      assert(file->has_hash());
-      Hash hash(file->get_hash());
-      if (!blob_cache->retrieve_file(blob_fds[next_blob_fd_idx++], path, false)) {
-        /* The file may not be writable but it may be expected and already checked. */
-        const FBBSTORE_Serialized_file* input_file =
-            find_input_file(
-                reinterpret_cast<const FBBSTORE_Serialized_process_inputs *>(inouts->get_inputs()),
-                path);
-        if (errno == EACCES && input_file && (file_to_file_info(file).mode_mask() & 0200)) {
-          /* The file has already been checked to be not writable and should be completely replaced
-           * from the cache. Let's remove it and try again. */
-          if (unlink(path->c_str()) == -1) {
-            fb_perror("Failed removing file to be replaced from cache");
+      /* Check if data is inlined */
+      fbb_size_t inline_data_len = file->get_inline_data_count();
+      if (inline_data_len > 0) {
+        FB_DEBUG(FB_DEBUG_SHORTCUT,
+                 "│   Restoring file from inline data: "
+                 + d(path) + " size=" + d(inline_data_len));
+        const char *inline_data = file->get_inline_data();
+
+        /* Write inline data to file */
+        int fd = open(path->c_str(), O_WRONLY | O_CREAT | O_TRUNC, 0644);
+        if (fd < 0) {
+          FB_DEBUG(FB_DEBUG_SHORTCUT, "│   Could not open file for writing: " + d(path));
+          return false;
+        }
+        ssize_t written = 0;
+        while (std::cmp_less(written, inline_data_len)) {
+          ssize_t n = write(fd, inline_data + written, inline_data_len - written);
+          if (n <= 0) {
+            close(fd);
+            FB_DEBUG(FB_DEBUG_SHORTCUT, "│   Could not write inline data to file: " + d(path));
+            return false;
+          }
+          written += n;
+        }
+        close(fd);
+
+        /* Set mode if specified */
+        if (file->has_mode()) {
+          chmod(path->c_str(), file->get_mode());
+        }
+      } else {
+        FB_DEBUG(FB_DEBUG_SHORTCUT,
+                 "│   Fetching file from blobs cache: "
+                 + d(path));
+        if (!blob_cache->retrieve_file(blob_fds[next_blob_fd_idx++], path, false)) {
+          /* The file may not be writable but it may be expected and already checked. */
+          const FBBSTORE_Serialized_file* input_file =
+              find_input_file(
+                  reinterpret_cast<const FBBSTORE_Serialized_process_inputs *>(
+                      inouts->get_inputs()), path);
+          if (errno == EACCES && input_file && (file_to_file_info(file).mode_mask() & 0200)) {
+            /* The file has already been checked to be not writable and should be completely replaced
+             * from the cache. Let's remove it and try again. */
+            if (unlink(path->c_str()) == -1) {
+              fb_perror("Failed removing file to be replaced from cache");
+              assert(0);
+            }
+            /* Try retrieving the same file again. */
+            if (!blob_cache->retrieve_file(blob_fds[next_blob_fd_idx - 1], path, false)) {
+              fb_perror("Failed creating file from cache");
+              assert(0);
+            }
+          } else {
+            fb_perror("Failed opening file to be recreated from cache");
             assert(0);
           }
-          /* Try retrieving the same file again. */
-          if (!blob_cache->retrieve_file(blob_fds[next_blob_fd_idx - 1], path, false)) {
-            fb_perror("Failed creating file from cache");
-            assert(0);
-          }
-        } else {
-          fb_perror("Failed opening file to be recreated from cache");
-          assert(0);
+        }
+        if (file->has_mode()) {
+          /* Refuse to apply setuid, setgid, sticky bit. */
+          // FIXME warn on them, even when we store them.
+          chmod(path->c_str(), file->get_mode() & 0777);
         }
       }
     }
-    if (file->has_mode()) {
-      /* Refuse to apply setuid, setgid, sticky bit. */
-      // FIXME warn on them, even when we store them.
-      chmod(path->c_str(), file->get_mode() & 0777);
-    }
     if (registration_point) {
       FileUsageUpdate update = file_to_file_usage_update(path, file);
       registration_point->register_file_usage_update(path, update);
@@ -1501,28 +1572,63 @@ bool ExecedProcessCacher::apply_shortcut(ExecedProcess *proc,
       Pipe *pipe = ffd->pipe().get();
       assert(pipe);
 
-      Hash hash(serialized_append_to_fd->get_hash());
-      int fd = blob_fds[next_blob_fd_idx++];
-      struct stat64 st;
-      if (fstat64(fd, &st) < 0) {
-        assert(0 && "fstat");
-      }
-      pipe->add_data_from_fd(fd, st.st_size);
+      /* Check if data is inlined */
+      fbb_size_t inline_data_len = serialized_append_to_fd->get_inline_data_count();
 
-      if (proc->parent()) {
-        /* Bubble up the replayed pipe data. */
-        std::vector<std::shared_ptr<PipeRecorder>>& recorders =
-            pipe->proc2recorders[proc->parent_exec_point()];
-        PipeRecorder::record_data_from_regular_fd(&recorders, fd, st.st_size);
+      if (inline_data_len > 0) {
+        const char *inline_data = serialized_append_to_fd->get_inline_data();
+        pipe->add_data_from_buffer(inline_data, inline_data_len);
+        /* Also bubble up the replayed pipe data. */
+        if (proc->parent()) {
+          /* Bubble up the replayed pipe data from inline buffer. */
+          std::vector<std::shared_ptr<PipeRecorder>>& recorders =
+              pipe->proc2recorders[proc->parent_exec_point()];
+          PipeRecorder::record_data_from_buffer(&recorders, inline_data, inline_data_len);
+        }
+      } else {
+        /* Data is in blob cache, use fd */
+        int fd = blob_fds[next_blob_fd_idx++];
+        struct stat64 st;
+        if (fstat64(fd, &st) < 0) {
+          assert(0 && "fstat");
+        }
+        pipe->add_data_from_fd(fd, st.st_size);
+
+        if (proc->parent()) {
+          /* Bubble up the replayed pipe data. */
+          std::vector<std::shared_ptr<PipeRecorder>>& recorders =
+              pipe->proc2recorders[proc->parent_exec_point()];
+          PipeRecorder::record_data_from_regular_fd(&recorders, fd, st.st_size);
+        }
       }
     } else if (ffd->type() == FD_FILE) {
       assert(ffd->filename());
 
-      FB_DEBUG(FB_DEBUG_SHORTCUT,
-               "│   Fetching file fragment from blobs cache: "
-               + d(ffd->filename()));
-      Hash hash(serialized_append_to_fd->get_hash());
-      blob_cache->retrieve_file(blob_fds[next_blob_fd_idx++], ffd->filename(), true);
+      /* Check if data is inlined */
+      fbb_size_t inline_data_len = serialized_append_to_fd->get_inline_data_count();
+
+      if (inline_data_len > 0) {
+        const char *inline_data = serialized_append_to_fd->get_inline_data();
+        /* Write inline data to file */
+        FB_DEBUG(FB_DEBUG_SHORTCUT,
+                 "│   Writing inline data to file: "
+                 + d(ffd->filename()));
+        int fd = open(ffd->filename()->c_str(), O_WRONLY | O_APPEND | O_CREAT, 0644);
+        if (fd >= 0) {
+          fb_write(fd, inline_data, inline_data_len);
+          close(fd);
+        } else {
+          fb_perror("open for inline data write");
+          assert(0);
+        }
+      } else {
+        /* Data is in blob cache */
+        FB_DEBUG(FB_DEBUG_SHORTCUT,
+                 "│   Fetching file fragment from blobs cache: "
+                 + d(ffd->filename()));
+        Hash hash(serialized_append_to_fd->get_hash());
+        blob_cache->retrieve_file(blob_fds[next_blob_fd_idx++], ffd->filename(), true);
+      }
 
       /* Tell the interceptor to seek forward in this fd. */
       fds_appended_to->push_back(append_to_fd);
@@ -1595,7 +1701,7 @@ bool ExecedProcessCacher::shortcut(ExecedProcess *proc, std::vector<int> *fds_ap
     }
     /* Trigger cleanup of ProcessInputsOutputs. */
     inouts = nullptr;
-    munmap(inouts_buf, inouts_buf_len);
+    ObjCache::unmap_entry(inouts_buf, inouts_buf_len);
   }
   FB_DEBUG(FB_DEBUG_SHORTCUT, "└─");
 
@@ -1661,7 +1767,7 @@ bool ExecedProcessCacher::is_entry_usable(uint8_t* entry_buf,
       reinterpret_cast<const FBBSTORE_Serialized_process_outputs *>(inouts->get_outputs());
   for (size_t i = 0; i < outputs->get_path_isreg_count(); i++) {
     auto file = reinterpret_cast<const FBBSTORE_Serialized_file *>(outputs->get_path_isreg_at(i));
-    if (file->get_type() == ISREG && file->has_hash()) {
+    if (file->get_type() == ISREG && file->get_inline_data_count() == 0 && file->has_hash()) {
       if (!blob_present(Hash(file->get_hash()), referenced_blobs)) {
         return false;
       }
@@ -1670,7 +1776,8 @@ bool ExecedProcessCacher::is_entry_usable(uint8_t* entry_buf,
   for (size_t i = 0; i < outputs->get_append_to_fd_count(); i++) {
     auto append_to_fd = reinterpret_cast<const FBBSTORE_Serialized_append_to_fd *>
         (outputs->get_append_to_fd_at(i));
-    if (!blob_present(Hash(append_to_fd->get_hash()), referenced_blobs)) {
+    if ((append_to_fd->get_inline_data_count() == 0)
+        && !blob_present(Hash(append_to_fd->get_hash()), referenced_blobs)) {
       return false;
     }
   }
