@@ -161,15 +161,20 @@ class Pipe;
 class PipeRecorder {
  public:
   explicit PipeRecorder(const ExecedProcess *for_proc);
-  ~PipeRecorder() { if (!abandoned_) abandon(); }
+  ~PipeRecorder() {
+    if (!abandoned_) abandon();
+    free(mem_buffer_);
+  }
 
   /**
    * If there was traffic, store it in the cache and return is_empty_out=false, key_out=[the_hash].
    * If there was no traffic, set is_empty_out=true, key_out is undefined.
+   * If the data is small enough (<=max_inline_blob_size), it is returned in inline_data_out.
    * Set the recorder to abandoned state.
    * Returns false in case of failure.
    */
-  bool store(bool *is_empty_out, Hash *key_out, off_t* stored_bytes);
+  bool store(bool *is_empty_out, Hash *key_out, off_t* stored_bytes,
+             char **inline_data_out, size_t *inline_data_len_out);
   /** Close the backing fd, drop the data that was written so far. Set to deactivated state. */
   void deactivate();
   /** Close the backing fd, drop the data that was written so far. Set to abandoned state. */
@@ -229,6 +234,15 @@ class PipeRecorder {
    * To be called the first time when there's data to record.
    */
   void open_backing_file();
+  /**
+   * Prepare for appending len bytes: decide between in-memory buffering and file-backed storage.
+   * If staying in memory, ensures capacity is available. If switching to file, opens the backing
+   * file if needed and flushes any previously buffered memory content to the file.
+   *
+   * @return true if the caller should append into the in-memory buffer, false if it should append
+   *         into the backing file descriptor.
+   */
+  bool prepare_for_append(ssize_t len);
   /**
    * Add non-empty data to this PipeRecorder from a memory buffer, using write().
    *
@@ -265,6 +279,13 @@ class PipeRecorder {
   /* The amount of data written so far. */
   loff_t offset_ {0};
 
+  /* In-memory buffer for small blobs (when size <= max_inline_blob_size) */
+  char *mem_buffer_ = NULL;
+  /* Size of allocated memory buffer */
+  size_t mem_buffer_capacity_ {0};
+  /* Whether we're buffering data in memory (true until size exceeds threshold) */
+  bool use_memory_buffer_ {true};
+
   /* A deactivated PipeRecorder has thrown away all the data it has ever received, and won't record
    * any new data. However, it might still receive data "to record", which will silently be dropped.
    * This is useful when a process can no longer be shortcut, so there's no point in recording the
